@@ -35,18 +35,18 @@ import logging
 log = logging.getLogger("braille")
 
 import signal
+import os
+
+import gobject
+gobject.threads_init()
 
 try:
     import louis
 except ImportError:
     louis = None
-    _defaultContractionTable = None
-else:
-    _defaultContractionTable = louis.getDefaultTable()
 
 try:
     import brlapi
-    import gobject
 
     _brlAPI = None
     _brlAPIAvailable = True
@@ -70,15 +70,39 @@ import eventsynthesizer
 import orca_state
 import settings
 
-from orca_i18n import _                          # for gettext support
-
-# If True, this module has been initialized.
+# Right now, the orca autogen.sh/configure needs a priori knowledge of
+# where the liblouis tables are.  When running autogen.sh/configure,
+# platform.py:tablesdir will be set to point to the liblouis table
+# location.  If not found, it will be the empty string.  We need to
+# capture that error condition, otherwise braille contraction will
+# just plain fail.  See also bgo#610134.  [[TODO: WDW - see if the
+# liblouis bindings can give us the tablesdir information at runtime
+# http://code.google.com/p/liblouis/issues/detail?id=9]]
 #
-_initialized = False
+from platform import tablesdir
+if louis and not tablesdir:
+    debug.println(debug.LEVEL_SEVERE,
+                  "Contraction tables for liblouis cannot be found.")
+    debug.println(debug.LEVEL_SEVERE,
+                  "This usually means orca was built before")
+    debug.println(debug.LEVEL_SEVERE,
+                  "liblouis was installed. Contracted braille will")
+    debug.println(debug.LEVEL_SEVERE,
+                  "not be available.")
+    louis = None
+    
+from orca_i18n import _                          # for gettext support
 
 # The braille monitor
 #
-monitor = None
+_monitor = None
+
+# brlapi keys which are not allowed to interupt speech:
+#
+dontInteruptSpeechKeys = []
+if _brlAPIAvailable:
+    dontInteruptSpeechKeys = [brlapi.KEY_CMD_FWINLT, brlapi.KEY_CMD_FWINRT, \
+        brlapi.KEY_CMD_LNUP, brlapi.KEY_CMD_LNDN]
 
 # Common names for most used BrlTTY commands, to be shown in the GUI:
 # ATM, the ones used in default.py are:
@@ -179,7 +203,8 @@ if _brlAPIAvailable:
 #
 # [[[TODO: WDW - Only a height of 1 is support at this time.]]]
 #
-_displaySize = [32, 1]
+DEFAULT_DISPLAY_SIZE = 32
+_displaySize = [DEFAULT_DISPLAY_SIZE, 1]
 
 # The list of lines on the display.  This represents the entire amount of data
 # to be drawn on the display.  It will be clipped by the viewport if too large.
@@ -190,6 +215,15 @@ _lines = []
 #
 _regionWithFocus = None
 
+# The last text information painted.  This has the following fields:
+#
+# lastTextObj = the last accessible
+# lastCaretOffset = the last caret offset of the last text displayed
+# lastLineOffset = the last line offset of the last text displayed
+# lastCursorCell = the last cell on the braille display for the caret
+#
+_lastTextInfo = (None, 0, 0, 0)
+
 # The viewport is a rectangular region of size _displaySize whose upper left
 # corner is defined by the point (x, line number).  As such, the viewport is
 # identified solely by its upper left point.
@@ -212,6 +246,75 @@ beginningIsShowing = False
 #
 cursorCell = 0
 
+# The event source of a timeout used for flashing a message.
+#
+_flashEventSourceId = 0
+
+# Line information saved prior to flashing any messages
+#
+_saved = None
+
+# Translators: These are the braille translation table names for different
+# languages. You could read about braille tables at:
+# http://en.wikipedia.org/wiki/Braille
+#
+TABLE_NAMES = {"Cz-Cz-g1": _("Czech Grade 1"),
+               "Es-Es-g1": _("Spanish Grade 1"),
+               "Fr-Ca-g2": _("Canada French Grade 2"),
+               "Fr-Fr-g2": _("France French Grade 2"),
+               "Lv-Lv-g1": _("Latvian Grade 1"),
+               "Nl-Nl-g1": _("Netherlands Dutch Grade 1"),
+               "No-No-g0": _("Norwegian Grade 0"),
+               "No-No-g1": _("Norwegian Grade 1"),
+               "No-No-g2": _("Norwegian Grade 2"),
+               "No-No-g3": _("Norwegian Grade 3"),
+               "Pl-Pl-g1": _("Polish Grade 1"),
+               "Pt-Pt-g1": _("Portuguese Grade 1"),
+               "Se-Se-g1": _("Swedish Grade 1"),
+               "ar-ar-g1": _("Arabic Grade 1"),
+               "cy-cy-g1": _("Welsh Grade 1"),
+               "cy-cy-g2": _("Welsh Grade 2"),
+               "de-de-g0": _("German Grade 0"),
+               "de-de-g1": _("German Grade 1"),
+               "de-de-g2": _("German Grade 2"),
+               "en-GB-g2": _("U.K. English Grade 2"),
+               "en-gb-g1": _("U.K. English Grade 1"),
+               "en-us-g1": _("U.S. English Grade 1"),
+               "en-us-g2": _("U.S. English Grade 2"),
+               "fr-ca-g1": _("Canada French Grade 1"),
+               "fr-fr-g1": _("France French Grade 1"),
+               "gr-gr-g1": _("Greek Grade 1"),
+               "hi-in-g1": _("Hindi Grade 1"),
+               "it-it-g1": _("Italian Grade 1"),
+               "nl-be-g1": _("Belgium Dutch Grade 1")}
+
+def listTables():
+    tables = {}
+    try:
+        for fname in os.listdir(tablesdir):
+            if fname[-4:] in (".utb", ".ctb"):
+                alias = fname[:-4]
+                tables[TABLE_NAMES.get(alias, alias)] = \
+                    os.path.join(tablesdir, fname)
+    except OSError:
+        pass
+
+    return tables
+
+def getDefaultTable():
+    try:
+        for fname in os.listdir(tablesdir):
+            if fname[-4:] in (".utb", ".ctb"):
+                if fname.startswith("en-us"):
+                    return os.path.join(tablesdir, fname)
+    except OSError:
+        pass
+
+    return ""
+
+if louis:
+    _defaultContractionTable = getDefaultTable()
+
 def _printBrailleEvent(level, command):
     """Prints out a Braille event.  The given level may be overridden
     if the eventDebugLevel (see debug.setEventDebugLevel) is greater in
@@ -263,6 +366,9 @@ class Region:
             self.string = self.rawLine
             self.cursorOffset = cursorOffset
 
+    def __str__(self):
+        return "Region: '%s', %d" % (self.string, self.cursorOffset)
+
     def processRoutingKey(self, offset):
         """Processes a cursor routing key press on this Component.  The offset
         is 0-based, where 0 represents the leftmost character of string
@@ -316,16 +422,19 @@ class Region:
             cursorOnSpace = False
 
         if not expandOnCursor or cursorOnSpace:
-            contracted, inPos, outPos, cursorPos = \
-                             louis.translate([self.contractionTable],
-                                             line,
-                                             cursorPos=cursorOffset)
+            mode = 0
         else:
-            contracted, inPos, outPos, cursorPos = \
-                             louis.translate([self.contractionTable],
-                                             line,
-                                             cursorPos=cursorOffset,
-                                             mode=louis.MODE.compbrlAtCursor)
+            mode = louis.compbrlAtCursor
+
+        contracted, inPos, outPos, cursorPos = \
+            louis.translate([self.contractionTable],
+                            line,
+                            cursorPos=cursorOffset,
+                            mode=mode)
+
+        # Make sure the cursor is at a realistic spot.
+        #
+        cursorPos = min(cursorPos, len(contracted))
 
         return contracted, inPos, outPos, cursorPos
 
@@ -394,6 +503,9 @@ class Component(Region):
 
         self.accessible = accessible
 
+    def __str__(self):
+        return "Component: '%s', %d" % (self.string, self.cursorOffset)
+
     def getCaretOffset(self, offset):
         """Returns the caret position of the given offset if the object
         has text with a caret.  Otherwise, returns -1.
@@ -421,6 +533,8 @@ class Component(Region):
             try:
                 eventsynthesizer.clickObject(self.accessible, 1)
             except:
+                debug.println(debug.LEVEL_SEVERE,
+                              "Could not process routing key:")
                 debug.printException(debug.LEVEL_SEVERE)
         else:
             action.doAction(0)
@@ -435,6 +549,9 @@ class Link(Component):
         have the region expand on cursor."""
         Component.__init__(self, accessible, string, cursorOffset, '', True)
 
+    def __str__(self):
+        return "Link: '%s', %d" % (self.string, self.cursorOffset)
+
     def getAttributeMask(self, getLinkMask=True):
         """Creates a string which can be used as the attrOr field of brltty's
         write structure for the purpose of indicating text attributes and
@@ -469,12 +586,14 @@ class Text(Region):
         """
 
         self.accessible = accessible
-        if orca_state.activeScript:
+        if orca_state.activeScript and self.accessible:
             [string, self.caretOffset, self.lineOffset] = \
                  orca_state.activeScript.getTextLineAtCaret(self.accessible,
                                                             startOffset)
         else:
             string = ""
+            self.caretOffset = 0
+            self.lineOffset = 0
 
         string = string.decode("UTF-8")
         if label:
@@ -515,6 +634,14 @@ class Text(Region):
 
         if not self.contracted and not settings.disableBrailleEOL:
             self.string += self.eol
+        elif settings.disableBrailleEOL:
+            # Ensure there is a place to click on at the end of a line
+            # so the user can route the caret to the end of the line.
+            #
+            self.string += ' '
+
+    def __str__(self):
+        return "Text: '%s', %d" % (self.string, self.cursorOffset)
 
     def repositionCursor(self):
         """Attempts to reposition the cursor in response to a new
@@ -572,7 +699,8 @@ class Text(Region):
         if caretOffset < 0:
             return
 
-        orca_state.activeScript.setCaretOffset(self.accessible, caretOffset)
+        orca_state.activeScript.utilities.setCaretOffset(
+            self.accessible, caretOffset)
 
     def getAttributeMask(self, getLinkMask=True):
         """Creates a string which can be used as the attrOr field of brltty's
@@ -621,14 +749,14 @@ class Text(Region):
                 n += 1
 
         if attrIndicator:
-            enabledAttributes = script.attributeStringToDictionary(
+            keys, enabledAttributes = script.utilities.stringToKeysAndDict(
                 settings.enabledBrailledTextAttributes)
 
             offset = self.lineOffset
             while offset < lineEndOffset:
                 attributes, startOffset, endOffset = \
-                            script.getTextAttributes(self.accessible,
-                                                     offset, True)
+                    script.utilities.textAttributes(self.accessible,
+                                                    offset, True)
                 if endOffset <= offset:
                     break
                 mask = settings.TEXT_ATTR_BRAILLE_NONE
@@ -645,7 +773,7 @@ class Text(Region):
                         regionMask[i] |= attrIndicator
 
         if selIndicator:
-            selections = script.getAllTextSelections(self.accessible)
+            selections = script.utilities.allTextSelections(self.accessible)
             for startOffset, endOffset in selections:
                 maskStart = max(startOffset - self.lineOffset, 0)
                 maskEnd = min(endOffset - self.lineOffset, stringLength)
@@ -752,7 +880,8 @@ class ReviewText(Region):
         been scrolled off the display."""
 
         caretOffset = self.getCaretOffset(offset)
-        orca_state.activeScript.setCaretOffset(self.accessible, caretOffset)
+        orca_state.activeScript.utilities.setCaretOffset(
+            self.accessible, caretOffset)
 
 class Line:
     """A horizontal line on the display.  Each Line is composed of a sequential
@@ -969,7 +1098,130 @@ def setFocus(region, panToFocus=True, getLinkMask=True):
 
     viewport[0] = max(0, offset)
 
-def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
+def _realignViewport(string, focusOffset, cursorOffset):
+    """Realigns the braille display to account for braille alignment
+    preferences.  By the time this method is called, if there is a
+    cursor cell to be displayed, it should already be somewhere in
+    the viewport.  All we're going to do is adjust the viewport a
+    little to align the viewport edge according to the
+    settings.brailleAlignmentStyle.
+
+    Arguments:
+    - string: the entire string to be presented
+    - focusOffset: where in string the focused region begins
+    - cursorOffset: where in the string the cursor should be
+
+    Returns: the viewport[0] value is potentially modified.
+    """
+
+    # pylint complains we don't set viewport, which in fact we do if
+    # 'jump' ends up being set.
+    #
+    # pylint: disable-msg=W0602
+    #
+    global viewport
+
+    jump = 0
+
+    # If there's no cursor to show or we're doing
+    # ALIGN_BRAILLE_BY_EDGE, the viewport should already be where it
+    # belongs.  Otherwise, we may need to do some adjusting of the
+    # viewport.
+    #
+    if (cursorOffset < 0) \
+       or (settings.brailleAlignmentStyle == settings.ALIGN_BRAILLE_BY_EDGE) \
+       or not (cursorOffset >= viewport[0]
+               and cursorOffset < (viewport[0] + _displaySize[0])):
+        pass
+    else:
+        # The left and right margin values are absolute values in the
+        # string and represent where in the string the margins of the
+        # current viewport lie.  Note these are margins and not the
+        # actual edges of the viewport.
+        #
+        leftMargin = viewport[0] + settings.brailleAlignmentMargin - 1
+        rightMargin = (viewport[0] + _displaySize[0]) \
+                      - settings.brailleAlignmentMargin
+
+        # This represents how far left in the string we want to search
+        # and also how far left we'll realign the viewport. Setting it
+        # to focusOffset means we won't move the viewport further left
+        # than the beginning of the current region with focus.
+        #
+        leftMostEdge = max(0, focusOffset)
+
+        # If we align by margin, we just want to keep the cursor at or
+        # in between the margins.  The only time we go outside the
+        # margins are when we are at the ends of the string.
+        #
+        if settings.brailleAlignmentStyle == settings.ALIGN_BRAILLE_BY_MARGIN:
+            if cursorOffset < leftMargin:
+                jump = cursorOffset - leftMargin
+            elif cursorOffset > rightMargin:
+                jump = cursorOffset - rightMargin
+        elif settings.brailleAlignmentStyle == settings.ALIGN_BRAILLE_BY_WORD:
+            # When we align by word, we want to try to show complete
+            # words at the edges of the braille display.  When we're
+            # near the left edge, we'll try to start a word at the
+            # left edge.  When we're near the right edge, we'll try to
+            # end a word at the right edge.
+            #
+            if cursorOffset < leftMargin:
+                # Find the index of the character that is the first
+                # letter of the word prior to left edge of the
+                # viewport.
+                #
+                inWord = False
+                leftWordEdge = viewport[0] - 1
+                while leftWordEdge >= leftMostEdge:
+                    if not string[leftWordEdge] in ' \t\n\r\v\f':
+                        inWord = True
+                    elif inWord:
+                        leftWordEdge += 1
+                        break
+                    leftWordEdge -= 1
+                leftWordEdge = max(leftMostEdge, leftWordEdge)
+                jump = leftWordEdge - viewport[0]
+            elif cursorOffset > rightMargin:
+                # Find the index of the character that is the last
+                # letter of the word after the right edge of the
+                # viewport.
+                #
+                inWord = False
+                rightWordEdge = viewport[0] + _displaySize[0]
+                while rightWordEdge < len(string):
+                    if not string[rightWordEdge] in ' \t\n\r\v\f':
+                        inWord = True
+                    elif inWord:
+                        break
+                    rightWordEdge += 1
+                rightWordEdge = min(len(string), rightWordEdge)
+                jump = max(0, rightWordEdge - (viewport[0] + _displaySize[0]))
+
+            # We use the brailleMaximumJump to help us handle really
+            # long words.  The (jump/abs(jump)) stuff is a quick and
+            # dirty way to retain the sign (i.e., +1 or -1).
+            #
+            if abs(jump) > settings.brailleMaximumJump:
+                jump = settings.brailleMaximumJump * (jump/abs(jump))
+
+    if jump:
+        # Set the viewport's left edge based upon the jump, making
+        # sure we don't go any farther left than the leftMostEdge.
+        #
+        viewport[0] = max(leftMostEdge, viewport[0] + jump)
+
+        # Now, make sure we don't scroll too far to the right.  That
+        # is, avoid showing blank spaces to the right if there is more
+        # of the string that can be shown.
+        #
+        viewport[0] = min(viewport[0],
+                          max(leftMostEdge, len(string) - _displaySize[0]))
+
+def refresh(panToCursor=True,
+            targetCursorCell=0,
+            getLinkMask=True,
+            stopFlash=True):
     """Repaints the Braille on the physical display.  This clips the entire
     logical structure by the viewport and also sets the cursor to the
     appropriate location.  [[[TODO: WDW - I'm not sure how BrlTTY handles
@@ -990,16 +1242,44 @@ def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
       attributeMask for links. Reasons we might not want to include
       knowning that we will fail and/or it taking an unreasonable
       amount of time (AKA Gecko).
+    - stopFlash: if True, kill any flashed message that may be showing.
     """
 
     global endIsShowing
     global beginningIsShowing
     global cursorCell
-    global monitor
+    global _monitor
+    global _lastTextInfo
+
+    # Check out what we were displaying the last time - it might be
+    # the same text object we are displaying now.
+    #
+    (lastTextObj, lastCaretOffset, lastLineOffset, lastCursorCell) = \
+        _lastTextInfo
+    if _regionWithFocus and isinstance(_regionWithFocus, Text):
+        currentTextObj = _regionWithFocus.accessible
+        currentCaretOffset = _regionWithFocus.caretOffset
+        currentLineOffset = _regionWithFocus.lineOffset
+    else:
+        currentTextObj = None
+        currentCaretOffset = 0
+        currentLineOffset = 0
+
+    if stopFlash:
+        killFlash(restoreSaved=False)
 
     if len(_lines) == 0:
+        if not _brlAPIRunning:
+            init(_callback, settings.tty)
         if _brlAPIRunning:
-            _brlAPI.writeText("", 0)
+            try:
+                _brlAPI.writeText("", 0)
+            except:
+                debug.println(debug.LEVEL_WARNING,
+                              "BrlTTY seems to have disappeared:")
+                debug.printException(debug.LEVEL_WARNING)
+                shutdown()
+        _lastTextInfo = (None, 0, 0, 0)
         return
 
     # Now determine the location of the cursor.  First, we'll figure
@@ -1010,6 +1290,33 @@ def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
     if targetCursorCell < 0:
         targetCursorCell = _displaySize[0] + targetCursorCell + 1
 
+    # If there is no target cursor cell, then try to set one.  We
+    # currently only do this for text objects, and we do so by looking
+    # at the last position of the caret offset and cursor cell.  The
+    # primary goal here is to keep the cursor movement on the display
+    # somewhat predictable.
+    #
+    if (targetCursorCell == 0) \
+       and currentTextObj and (currentTextObj == lastTextObj) \
+       and (currentLineOffset == lastLineOffset):
+        if lastCursorCell == 0:
+            # The lastCursorCell will be 0 if the user has panned
+            # the display on a long line and the caret of the text
+            # object is no longer in view.  We'll pass here and
+            # let the panning code figure out what to do.
+            #
+            pass
+        elif lastCaretOffset == currentCaretOffset:
+            targetCursorCell = lastCursorCell
+        elif lastCaretOffset < currentCaretOffset:
+            targetCursorCell = min(_displaySize[0],
+                                   lastCursorCell \
+                                   + (currentCaretOffset - lastCaretOffset))
+        elif lastCaretOffset > currentCaretOffset:
+            targetCursorCell = max(1,
+                                   lastCursorCell \
+                                   - (lastCaretOffset - currentCaretOffset))
+
     # Now, we figure out the 0-based offset for where the cursor
     # actually is in the string.
     #
@@ -1034,6 +1341,12 @@ def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
         elif cursorOffset >= (viewport[0] + _displaySize[0]):
             viewport[0] = max(0, cursorOffset - _displaySize[0] + 1)
 
+    # The cursorOffset should be somewhere in the viewport right now.
+    # Let's try to realign the viewport so that the cursor shows up
+    # according to the settings.brailleAlignmentStyle setting.
+    #
+    _realignViewport(string, focusOffset, cursorOffset)
+
     startPos = viewport[0]
     endPos = startPos + _displaySize[0]
 
@@ -1055,6 +1368,8 @@ def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
     log.info(logLine.encode("UTF-8"))
 
     substring = string[startPos:endPos]
+    if not _brlAPIRunning:
+        init(_callback, settings.tty)
     if _brlAPIRunning:
         writeStruct = brlapi.WriteStruct()
         writeStruct.regionBegin = 1
@@ -1085,35 +1400,120 @@ def refresh(panToCursor=True, targetCursorCell=0, getLinkMask=True):
         if attributeMask:
             writeStruct.attrOr = attributeMask[startPos:endPos]
 
-        _brlAPI.write(writeStruct)
+        if not _brlAPIRunning:
+            init(_callback, settings.tty)
+        if _brlAPIRunning:
+            try:
+                _brlAPI.write(writeStruct)
+            except:
+                debug.println(debug.LEVEL_WARNING,
+                              "BrlTTY seems to have disappeared:")
+                debug.printException(debug.LEVEL_WARNING)
+                shutdown()
 
     if settings.enableBrailleMonitor:
-        if not monitor:
-            monitor = brlmon.BrlMon(_displaySize[0])
-            monitor.show_all()
+        if not _monitor:
+            _monitor = brlmon.BrlMon(_displaySize[0])
+            _monitor.show_all()
         if attributeMask:
             subMask = attributeMask[startPos:endPos]
         else:
             subMask = None
-        monitor.writeText(cursorCell, substring, subMask)
-    elif monitor:
-        monitor.destroy()
-        monitor = None
+        _monitor.writeText(cursorCell, substring, subMask)
+    elif _monitor:
+        _monitor.destroy()
+        _monitor = None
 
     beginningIsShowing = startPos == 0
     endIsShowing = endPos >= len(string)
 
-def displayRegions(regionInfo):
+    # Remember the text information we were presenting (if any)
+    #
+    if _regionWithFocus and isinstance(_regionWithFocus, Text):
+        _lastTextInfo = (_regionWithFocus.accessible,
+                         _regionWithFocus.caretOffset,
+                         _regionWithFocus.lineOffset,
+                         cursorCell)
+    else:
+        _lastTextInfo = (None, 0, 0, 0)
+
+def _flashCallback():
+    global _lines
+    global _regionWithFocus
+    global viewport
+    global _flashEventSourceId
+
+    if _flashEventSourceId:
+        (_lines, _regionWithFocus, viewport, flashTime) = _saved
+        refresh(panToCursor=False, stopFlash=False)
+        _flashEventSourceId = 0
+
+    return False
+
+def killFlash(restoreSaved=True):
+    global _flashEventSourceId
+    global _lines
+    global _regionWithFocus
+    global viewport
+    if _flashEventSourceId:
+        if _flashEventSourceId > 0:
+            gobject.source_remove(_flashEventSourceId)
+        if restoreSaved:
+            (_lines, _regionWithFocus, viewport, flashTime) = _saved
+            refresh(panToCursor=False, stopFlash=False)
+        _flashEventSourceId = 0
+
+def resetFlashTimer():
+    global _flashEventSourceId
+    if _flashEventSourceId > 0:
+        gobject.source_remove(_flashEventSourceId)
+        flashTime = _saved[3]
+        _flashEventSourceId = gobject.timeout_add(flashTime, _flashCallback)
+
+def _initFlash(flashTime):
+    """Sets up the state needed to flash a message or clears any existing
+    flash if nothing is to be flashed.
+
+    Arguments:
+    - flashTime:  if non-0, the number of milliseconds to display the
+                  regions before reverting back to what was there before.
+                  A 0 means to not do any flashing.  A negative number
+                  means display the message until some other message
+                  comes along or the user presses a cursor routing key.
+    """
+
+    global _saved
+    global _flashEventSourceId
+
+    if _flashEventSourceId:
+        if _flashEventSourceId > 0:
+            gobject.source_remove(_flashEventSourceId)
+        _flashEventSourceId = 0
+    else:
+        _saved = (_lines, _regionWithFocus, viewport, flashTime)
+
+    if flashTime > 0:
+        _flashEventSourceId = gobject.timeout_add(flashTime, _flashCallback)
+    elif flashTime < 0:
+        _flashEventSourceId = -666
+
+def displayRegions(regionInfo, flashTime=0):
     """Displays a list of regions on a single line, setting focus to the
        specified region.  The regionInfo parameter is something that is
-       typically returned by a call to braillegenerator.getBrailleRegions.
+       typically returned by a call to braille_generator.generateBraille.
 
     Arguments:
     - regionInfo: a list where the first element is a list of regions
                   to display and the second element is the region
                   with focus (must be in the list from element 0)
+    - flashTime:  if non-0, the number of milliseconds to display the
+                  regions before reverting back to what was there before.
+                  A 0 means to not do any flashing.  A negative number
+                  means display the message until some other message
+                  comes along or the user presses a cursor routing key.
     """
 
+    _initFlash(flashTime)
     regions = regionInfo[0]
     focusedRegion = regionInfo[1]
 
@@ -1123,22 +1523,28 @@ def displayRegions(regionInfo):
         line.addRegion(item)
     addLine(line)
     setFocus(focusedRegion)
-    refresh()
+    refresh(stopFlash=False)
 
-def displayMessage(message, cursor=-1):
+def displayMessage(message, cursor=-1, flashTime=0):
     """Displays a single line, setting the cursor to the given position,
     ensuring that the cursor is in view.
 
     Arguments:
     - message: the string to display
     - cursor: the 0-based cursor position, where -1 (default) means no cursor
+    - flashTime:  if non-0, the number of milliseconds to display the
+                  regions before reverting back to what was there before.
+                  A 0 means to not do any flashing.  A negative number
+                  means display the message until some other message
+                  comes along or the user presses a cursor routing key.
     """
 
+    _initFlash(flashTime)
     clear()
     region = Region(message, cursor)
     addLine(Line(region))
     setFocus(region)
-    refresh(True)
+    refresh(True, stopFlash=False)
 
 def panLeft(panAmount=0):
     """Pans the display to the left, limiting the pan to the beginning
@@ -1238,6 +1644,12 @@ def processRoutingKey(event):
     the dictionary form of the expanded BrlAPI event.
     """
 
+    # If a message is being flashed, we'll use a routing key to dismiss it.
+    #
+    if _flashEventSourceId:
+        killFlash()
+        return
+
     cell = event.event["argument"]
 
     if len(_lines) > 0:
@@ -1270,6 +1682,7 @@ def _processBrailleEvent(event):
             #
             consumed = _callback(event)
         except:
+            debug.println(debug.LEVEL_WARNING, "Issue processing event:")
             debug.printException(debug.LEVEL_WARNING)
             consumed = False
 
@@ -1282,7 +1695,13 @@ def _brlAPIKeyReader(source, condition):
     """Method to read a key from the BrlAPI bindings.  This is a
     gobject IO watch handler.
     """
-    key = _brlAPI.readKey(False)
+    try:
+        key = _brlAPI.readKey(False)
+    except:
+        debug.println(debug.LEVEL_WARNING, "BrlTTY seems to have disappeared:")
+        debug.printException(debug.LEVEL_WARNING)
+        shutdown()
+        return
     if key:
         _processBrailleEvent(_brlAPI.expandKeyCode(key))
     return _brlAPIRunning
@@ -1294,6 +1713,8 @@ def setupKeyRanges(keys):
     Arguments:
     -keys: a list of BrlAPI commands.
     """
+    if not _brlAPIRunning:
+        init(_callback, settings.tty)
     if not _brlAPIRunning:
         return
 
@@ -1318,29 +1739,29 @@ def init(callback=None, tty=7):
     Arguments:
     - callback: the method to call with a BrlTTY input event.
     - tty: the tty port to take ownership of (default = 7)
-    Returns True if the initialization procedure was run or False if this
-    module has already been initialized.
+    Returns False if BrlTTY cannot be accessed or braille has
+    not been enabled.
     """
 
-    global _initialized
+    global _brlAPI
+    global _brlAPIRunning
+    global _brlAPISourceId
     global _displaySize
     global _callback
+    global _monitor
 
-    if _initialized:
+    if _brlAPIRunning:
+        return True
+
+    if not settings.enableBraille:
         return False
 
     _callback = callback
 
     try:
-        global _brlAPI
-        global _brlAPIRunning
-        global _brlAPISourceId
-
-        gobject.threads_init()
         _brlAPI = brlapi.Connection()
 
         try:
-            import os
             windowPath = os.environ["WINDOWPATH"]
             _brlAPI.enterTtyModeWithPath()
             _brlAPIRunning = True
@@ -1358,7 +1779,10 @@ def init(callback=None, tty=7):
                                                gobject.IO_IN,
                                                _brlAPIKeyReader)
     except:
-        debug.printException(debug.LEVEL_FINEST)
+        debug.println(debug.LEVEL_CONFIGURATION,
+                      "Could not initialize BrlTTY:")
+        debug.printException(debug.LEVEL_CONFIGURATION)
+        _brlAPIRunning = False
         return False
 
     # [[[TODO: WDW - For some reason, BrlTTY wants to say the height of the
@@ -1368,6 +1792,12 @@ def init(callback=None, tty=7):
     (x, y) = _brlAPI.displaySize
     _displaySize = [x, 1]
 
+    # The monitor will be created in refresh if needed.
+    #
+    if _monitor:
+        _monitor.destroy()
+        _monitor = None
+
     debug.println(debug.LEVEL_CONFIGURATION,
                   "braille display size = (%d, %d)" \
                   % (_displaySize[0], _displaySize[1]))
@@ -1375,29 +1805,31 @@ def init(callback=None, tty=7):
     clear()
     refresh(True)
 
-    _initialized = True
-
     return True
 
 def shutdown():
     """Shuts down the braille module.   Returns True if the shutdown procedure
-    was run or False if this module has not been initialized.
+    was run.
     """
 
-    global _initialized
-
-    if not _initialized:
-        return False
-
     global _brlAPIRunning
     global _brlAPISourceId
+    global _monitor
+    global _displaySize
 
     if _brlAPIRunning:
         _brlAPIRunning = False
         gobject.source_remove(_brlAPISourceId)
         _brlAPISourceId = 0
-        _brlAPI.leaveTtyMode()
-
-    _initialized = False
+        try:
+            _brlAPI.leaveTtyMode()
+        except:
+            pass
+        if _monitor:
+            _monitor.destroy()
+            _monitor = None
+        _displaySize = [DEFAULT_DISPLAY_SIZE, 1]
+    else:
+        return False
 
     return True
