@@ -1,6 +1,6 @@
 # Orca
 #
-# Copyright 2004-2008 Sun Microsystems Inc.
+# Copyright 2004-2009 Sun Microsystems Inc.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Library General Public
@@ -22,7 +22,7 @@
 __id__  = "$Id$"
 __version__   = "$Revision$"
 __date__      = "$Date$"
-__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
+__copyright__ = "Copyright (c) 2005-2009 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import gobject
@@ -147,7 +147,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         - script: the script.
         """
 
-        if settings.listenAllEvents: 
+        if settings.listenAllEvents:
             # We are always listening to everything.
             return
 
@@ -161,7 +161,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         Arguments:
         - script: the script.
         """
-        if settings.listenAllEvents: 
+        if settings.listenAllEvents:
             # We are always listening to everything.
             return
 
@@ -497,27 +497,24 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         #
         try:
             if event.source.getRole() == pyatspi.ROLE_TOOL_TIP:
-                # Check that it's okay to present tool tips. Always present
-                # tooltips initiated by the user pressing Control-F1 on the
-                # keyboard.
-                #
-                if isinstance(orca_state.lastInputEvent, \
-                              input_event.KeyboardEvent):
-                    if not orca_state.lastNonModifierKeyEvent.event_string \
-                                                                     == "F1":
-                        return
-
-                    # Mouse move events don't update orca_state.lastInputEvent
-                    # so it's possible the user accidentally nudged the
-                    # mouse and generated another tooltip event. If the
-                    # current time minus the last keyboard event time is
-                    # greater than 0.2 seconds, than just ignore this tooltip
-                    # event.
+                if settings.presentToolTips:
+                    pass
+                elif isinstance(orca_state.lastInputEvent, \
+                                input_event.KeyboardEvent) \
+                     and orca_state.lastNonModifierKeyEvent.event_string \
+                         == "F1":
+                    # Always present tooltips initiated by the user
+                    # pressing Control-F1 on the keyboard.  Mouse move
+                    # events don't update orca_state.lastInputEvents,
+                    # however, so it's possible the user accidentally
+                    # nudged the mouse after pressing F1 and generated
+                    # another tooltip event. If the current time minus
+                    # the last keyboard event time is greater than 0.5
+                    # seconds, than just ignore this tooltip event.
                     #
-                    currentTime =  time.time()
-                    if (currentTime - orca_state.lastInputEvent.time) > 0.2:
+                    if (time.time() - orca_state.lastInputEvent.time) > 0.5:
                         return
-                elif not settings.presentToolTips:
+                else:
                     return
         except:
             pass
@@ -535,6 +532,16 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     self._cleanupGarbage()
             return
 
+        # Clean up any flat review context so that Orca does not get
+        # confused (see bgo#609633)
+        #
+        if event.type.startswith("window:deactivate") \
+           and orca_state.activeScript \
+           and orca_state.activeScript.flatReviewContext \
+           and orca_state.activeScript.app == event.host_application:
+            orca_state.activeScript.drawOutline(-1, 0, 0, 0)
+            orca_state.activeScript.flatReviewContext = None
+
         try:
             # We don't want to touch a defunct object.  It's useless and it
             # can cause hangs.
@@ -545,10 +552,6 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     debug.println(debug.LEVEL_FINEST,
                                   "IGNORING DEFUNCT OBJECT")
                     if event.type.startswith("window:deactivate"):
-                        if orca_state.activeScript \
-                           and orca_state.activeScript.flatReviewContext:
-                            orca_state.activeScript.drawOutline(-1, 0, 0, 0)
-                            orca_state.activeScript.flatReviewContext = None
                         orca.setLocusOfFocus(event, None)
                         orca_state.activeWindow = None
                     return
@@ -560,6 +563,12 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     debug.printDetails(debug.LEVEL_FINEST, "    ", event.source)
 
         except LookupError:
+            if event.type.startswith("window:deactivate"):
+                debug.println(debug.LEVEL_FINEST,
+                              "LookupError in deactivate event; clearing focus")
+                orca.setLocusOfFocus(event, None)
+                orca_state.activeWindow = None
+            return
             debug.printException(debug.LEVEL_WARNING)
             debug.println(debug.LEVEL_WARNING,
                           "LookupError above while processing event: %s" %\
@@ -670,15 +679,16 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                                 and (orca_state.activeScript.app \
                                      != event.host_application)
 
-                    # One last check -- let's make sure the new script actually
-                    # has an active window.  See bug #568696.
+                    # One last check -- let's make sure the new script
+                    # thinks it should become active.
                     #
-                    if setNewActiveScript and event.host_application:
-                        setNewActiveScript = False
-                        for child in event.host_application:
-                            if child.getState().contains(pyatspi.STATE_ACTIVE):
-                                setNewActiveScript = True
-                                break
+                    if setNewActiveScript:
+                        theScript = \
+                            self.getScript(event.host_application \
+                                           or event.source.getApplication())
+                        setNewActiveScript = theScript.isActivatableEvent(event)
+                        if not reason and setNewActiveScript:
+                            reason = "script requested it"
 
                     if not reason and setNewActiveScript:
                         reason = "object received focus"
@@ -798,7 +808,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             # If the event doesn't have a source or that source is not marked
             # valid, then we don't care about this event. Just return.
             #
-            
+
             event = e
             if not event.source:
                 debug.println(debug.LEVEL_FINEST,
@@ -813,11 +823,12 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                 debug.println(debug.LEVEL_ALL,
                               "---------> QUEUEING EVENT %s" % e.type)
 
-            # Some toolkits (e.g., Java - see bug #531869) need to
+            # Some toolkits (e.g., Java - see bug #531869) may need to
             # have their events processed immediately.
             #
             try:
-                if event.host_application.toolkitName \
+                if len(settings.synchronousToolkits) > 0 and \
+                    event.host_application.toolkitName \
                     in settings.synchronousToolkits:
                     asyncMode = False
             except:
@@ -1084,7 +1095,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         self._registerEventListener("window:deactivate")
         self._registerEventListener("object:children-changed:remove")
 
-        win = orca_state.activeScript.findActiveWindow()
+        win = orca_state.activeScript.utilities.activeWindow()
         if win:
             # Generate a fake window activation event so the application
             # can tell the user about itself.
@@ -1097,6 +1108,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     self.detail1 = detail1
                     self.detail2 = detail2
                     self.any_data = any_data
+                    self.host_application = None
 
             class _FakeData:
                 def __init__(self):
@@ -1105,8 +1117,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     return None
 
             fe = _FakeEvent(win, "window:activate", 0, 0, _FakeData())
-            e = pyatspi.event.Event(fe)
-            self._enqueueEvent(e)
+            self._enqueueEvent(fe)
 
     def deactivate(self):
         """Called when this presentation manager is deactivated."""
