@@ -17,8 +17,7 @@
 # Free Software Foundation, Inc., Franklin Street, Fifth Floor,
 # Boston MA  02110-1301 USA.
 
-""" Custom script for Thunderbird 3.
-"""
+""" Custom script for Thunderbird 3."""
 
 __id__        = "$Id$"
 __version__   = "$Revision$"
@@ -32,7 +31,6 @@ import pyatspi
 import orca.orca as orca
 import orca.debug as debug
 import orca.default as default
-import orca.input_event as input_event
 import orca.orca_state as orca_state
 import orca.settings as settings
 import orca.speech as speech
@@ -40,7 +38,9 @@ import orca.scripts.toolkits.Gecko as Gecko
 
 from orca.orca_i18n import _
 
+from formatting import Formatting
 from speech_generator import SpeechGenerator
+from script_utilities import Utilities
 import script_settings
 
 ########################################################################
@@ -85,11 +85,20 @@ class Script(Gecko.Script):
 
         self.textArea = None
 
+    def getFormatting(self):
+        """Returns the formatting strings for this script."""
+        return Formatting(self)
+
     def getSpeechGenerator(self):
-        """Returns the speech generator for this script.
-        """
+        """Returns the speech generator for this script."""
+
         return SpeechGenerator(self)
 
+    def getUtilities(self):
+        """Returns the utilites for this script."""
+
+        return Utilities(self)
+
     def getAppPreferencesGUI(self):
         """Return a GtkVBox contain the application unique configuration
         GUI items for the current application.
@@ -102,6 +111,13 @@ class Script(Gecko.Script):
         gtk.ToggleButton.set_active(self.sayAllOnLoadCheckButton,
                                     script_settings.sayAllOnLoad)
 
+        # We need to maintain a separate setting for grabFocusOnAncestor
+        # because the version of Gecko used by the Thunderbird might be
+        # different from that used by Firefox. See bug 608149.
+        #
+        gtk.ToggleButton.set_active(self.grabFocusOnAncestorCheckButton,
+                                    script_settings.grabFocusOnAncestor)
+
         return vbox
 
     def setAppPreferences(self, prefs):
@@ -114,16 +130,20 @@ class Script(Gecko.Script):
 
         Gecko.Script.setAppPreferences(self, prefs)
 
-        # Write the Thunderbird specific setting.
+        # Write the Thunderbird specific settings.
         #
         prefix = "orca.scripts.apps.Thunderbird.script_settings"
         value = self.sayAllOnLoadCheckButton.get_active()
         prefs.writelines("%s.sayAllOnLoad = %s\n" % (prefix, value))
         script_settings.sayAllOnLoad = value
 
+        value = self.grabFocusOnAncestorCheckButton.get_active()
+        prefs.writelines("%s.grabFocusOnAncestor = %s\n" % (prefix, value))
+        script_settings.grabFocusOnAncestor = value
+
     def _debug(self, msg):
-        """ Convenience method for printing debug messages
-        """
+        """ Convenience method for printing debug messages"""
+
         debug.println(self.debugLevel, "Thunderbird.py: "+msg)
 
     def _isSpellCheckListItemFocus(self, event):
@@ -141,7 +161,7 @@ class Script(Gecko.Script):
                      pyatspi.ROLE_LIST, \
                      pyatspi.ROLE_DIALOG, \
                      pyatspi.ROLE_APPLICATION]
-        if self.isDesiredFocusedItem(event.source, rolesList):
+        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
             dialog = event.source.parent.parent
 
             # Translators: this is what the name of the spell checking
@@ -180,10 +200,8 @@ class Script(Gecko.Script):
         # based on the locusOfFocus other times Gecko is because of the
         # caret context.
         #
-        updatePosition = False
-        if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
-            string = orca_state.lastNonModifierKeyEvent.event_string
-            updatePosition = string in ["Page_Up", "Page_Down"]
+        lastKey, mods = self.utilities.lastKeyAndModifiers()
+        updatePosition = lastKey in ["Page_Up", "Page_Down"]
 
         # Unlike the unpredictable wild, wild web, odds are good that a
         # caret-moved event in a message composition window is valid. But
@@ -226,7 +244,7 @@ class Script(Gecko.Script):
         """
         obj = event.source
         parent = obj.parent
-        top = self.getTopLevel(obj)
+        top = self.utilities.topLevelObject(obj)
         consume = False
 
         # Clear the stored autocomplete string.
@@ -247,7 +265,7 @@ class Script(Gecko.Script):
         #
         if obj.getRole() == pyatspi.ROLE_TABLE_CELL:
             table = parent.queryTable()
-            row = table.getRowAtIndex(self.getCellIndex(obj))
+            row = table.getRowAtIndex(self.utilities.cellIndex(obj))
             for i in range(0, table.nColumns):
                 acc = table.getAccessibleAt(row, i)
                 if acc.name:
@@ -277,7 +295,7 @@ class Script(Gecko.Script):
                      pyatspi.ROLE_INTERNAL_FRAME,
                      pyatspi.ROLE_FRAME,
                      pyatspi.ROLE_APPLICATION]
-        if self.isDesiredFocusedItem(event.source, rolesList):
+        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
             self._debug("onFocus - message text area.")
 
             self.textArea = event.source
@@ -292,7 +310,7 @@ class Script(Gecko.Script):
             rolesList = [pyatspi.ROLE_ENTRY, \
                          pyatspi.ROLE_DIALOG, \
                          pyatspi.ROLE_APPLICATION]
-            if self.isDesiredFocusedItem(obj, rolesList):
+            if self.utilities.hasMatchingHierarchy(obj, rolesList):
                 dialog = obj.parent
 
                 # Translators: this is what the name of the spell checking
@@ -377,11 +395,10 @@ class Script(Gecko.Script):
                      pyatspi.ROLE_TREE_TABLE,
                      pyatspi.ROLE_SCROLL_PANE,
                      pyatspi.ROLE_SCROLL_PANE]
-        if self.isDesiredFocusedItem(event.source, rolesList):
-            if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
-                string = orca_state.lastNonModifierKeyEvent.event_string
-                if string == "Delete":
-                    oldLocusOfFocus = None
+        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
+            lastKey, mods = self.utilities.lastKeyAndModifiers()
+            if lastKey == "Delete":
+                oldLocusOfFocus = None
 
         # If the user has just deleted an open mail message, then we want to
         # try to speak the new name of the open mail message frame.
@@ -391,14 +408,13 @@ class Script(Gecko.Script):
                      pyatspi.ROLE_INTERNAL_FRAME, \
                      pyatspi.ROLE_FRAME, \
                      pyatspi.ROLE_APPLICATION]
-        if self.isDesiredFocusedItem(event.source, rolesList):
-            if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
-                string = orca_state.lastNonModifierKeyEvent.event_string
-                if string == "Delete":
-                    oldLocusOfFocus = None
-                    state = newLocusOfFocus.getState()
-                    if state.contains(pyatspi.STATE_DEFUNCT):
-                        newLocusOfFocus = event.source
+        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
+            lastKey, mods = self.utilities.lastKeyAndModifiers()
+            if lastKey == "Delete":
+                oldLocusOfFocus = None
+                state = newLocusOfFocus.getState()
+                if state.contains(pyatspi.STATE_DEFUNCT):
+                    newLocusOfFocus = event.source
 
         # Pass the event onto the parent class to be handled in the default way.
 
@@ -495,18 +511,17 @@ class Script(Gecko.Script):
         # we do when a new message window is opened. See bug #540039 for more
         # details.
         #
-        rolesList = [pyatspi.ROLE_DOCUMENT_FRAME, \
-                     pyatspi.ROLE_INTERNAL_FRAME, \
-                     pyatspi.ROLE_FRAME, \
+        rolesList = [pyatspi.ROLE_DOCUMENT_FRAME,
+                     pyatspi.ROLE_INTERNAL_FRAME,
+                     pyatspi.ROLE_FRAME,
                      pyatspi.ROLE_APPLICATION]
-        if self.isDesiredFocusedItem(event.source, rolesList):
-            if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
-                string = orca_state.lastNonModifierKeyEvent.event_string
-                if string == "Delete":
-                    speech.speak(obj.name)
-                    [obj, offset] = self.findFirstCaretContext(obj, 0)
-                    self.setCaretPosition(obj, offset)
-                    return
+        if self.utilities.hasMatchingHierarchy(event.source, rolesList):
+            lastKey, mods = self.utilities.lastKeyAndModifiers()
+            if lastKey == "Delete":
+                speech.speak(obj.name)
+                [obj, offset] = self.findFirstCaretContext(obj, 0)
+                self.setCaretPosition(obj, offset)
+                return
 
         # If we get a "object:property-change:accessible-name" event for 
         # the first item in the Suggestions lists for the spell checking
@@ -514,11 +529,11 @@ class Script(Gecko.Script):
         # will by the "Misspelled word:" label and the currently misspelled
         # word. See bug #535192 for more details.
         #
-        rolesList = [pyatspi.ROLE_LIST_ITEM, \
-                     pyatspi.ROLE_LIST, \
-                     pyatspi.ROLE_DIALOG, \
+        rolesList = [pyatspi.ROLE_LIST_ITEM,
+                     pyatspi.ROLE_LIST,
+                     pyatspi.ROLE_DIALOG,
                      pyatspi.ROLE_APPLICATION]
-        if self.isDesiredFocusedItem(obj, rolesList):
+        if self.utilities.hasMatchingHierarchy(obj, rolesList):
             dialog = obj.parent.parent
 
             # Translators: this is what the name of the spell checking 
@@ -528,7 +543,7 @@ class Script(Gecko.Script):
             #
             if dialog.name.startswith(_("Check Spelling")):
                 if obj.getIndexInParent() == 0:
-                    badWord = self.getDisplayedText(dialog[1])
+                    badWord = self.utilities.displayedText(dialog[1])
 
                     if self.textArea != None:
                         # If we have a handle to the Thunderbird message text
@@ -536,7 +551,7 @@ class Script(Gecko.Script):
                         # create a list of all the words found in them.
                         #
                         allTokens = []
-                        text = self.getText(self.textArea, 0, -1)
+                        text = self.utilities.substring(self.textArea, 0, -1)
                         tokens = text.split()
                         allTokens += tokens
                         self.speakMisspeltWord(allTokens, badWord)
@@ -597,6 +612,17 @@ class Script(Gecko.Script):
             self.sayAll(None)
         self._messageLoaded = False
 
+    def sayCharacter(self, obj):
+        """Speaks the character at the current caret position."""
+
+        if self.isEditableMessage(obj):
+            text = self.utilities.queryNonEmptyText(obj)
+            if text and text.caretOffset + 1 >= text.characterCount:
+                default.Script.sayCharacter(self, obj)
+                return
+
+        Gecko.Script.sayCharacter(self, obj)
+
     def getBottomOfFile(self):
         """Returns the object and last caret offset at the bottom of the
         document frame. Overridden here to handle editable messages.
@@ -613,24 +639,6 @@ class Script(Gecko.Script):
 
         return [obj, offset]
 
-    def getDocumentFrame(self):
-        """Returns the document frame that holds the content being shown.
-        Overridden here because multiple open messages are not arranged
-        in tabs like they are in Firefox."""
-
-        if self.inFindToolbar():
-            return Gecko.Script.getDocumentFrame(self)
-
-        obj = orca_state.locusOfFocus
-        while obj:
-            role = obj.getRole()
-            if role in [pyatspi.ROLE_DOCUMENT_FRAME, pyatspi.ROLE_EMBEDDED]:
-                return obj
-            else:
-                obj = obj.parent
-
-        return None
-
     def toggleFlatReviewMode(self, inputEvent=None):
         """Toggles between flat review mode and focus tracking mode."""
 
@@ -648,9 +656,8 @@ class Script(Gecko.Script):
         Returns True is this is something like the Subject: entry
         """
         result = obj and obj.getRole() == pyatspi.ROLE_ENTRY \
-            and None == self.getAncestor(obj,
-                                         [pyatspi.ROLE_DOCUMENT_FRAME],
-                                         [pyatspi.ROLE_FRAME])
+            and not self.utilities.ancestorWithRole(
+                obj, [pyatspi.ROLE_DOCUMENT_FRAME], [pyatspi.ROLE_FRAME])
         return result
 
     def isEditableMessage(self, obj):
